
import numpy as np

def _calculate_running_quantile(x, y, p, n_bins):
    """
    Calculate the quantile of y in bins of x
    """

    idx = np.argsort(x)
    x, y = x[idx], y[idx]

    dx = (x[-1] - x[0]) / n_bins
    x_out = np.linspace(x[0]+dx/2, x[-1]-dx/2, n_bins)

    y_out = np.zeros(x_out.shape)

    for i in range(len(x_out)):
        idx = np.nonzero((x >= x_out[i]-dx/2) & (x < x_out[i]+dx/2))[0]
        if len(idx) > 0:
            y_out[i] = np.percentile(y[idx], p)
        else:
            if i > 0:
                y_out[i] = y_out[i-1]
            else:
                y_out[i] = np.nan

    return x_out, y_out



# _filter_static_genes.py

__module_name__ = "_filter_static_genes.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# package imports #
# --------------- #
import licorice
import scipy.optimize
import matplotlib.pyplot as plt
import numpy as np
import vinplots

from ._calculate_running_quantile import _calculate_running_quantile

def _get_base_idx(base_idx, n_cells):

    if len(base_idx) == 0:
        base_idx = np.arange(n_cells)
    return base_idx


def _get_mu_gene(X, min_mean):

    mu_gene_ = X.mean(axis=0).A.squeeze()
    gene_idx = np.nonzero(mu_gene_ > min_mean)[0]
    mu_gene = mu_gene_[gene_idx]

    return mu_gene, gene_idx


def _get_variable_gene(X, min_mean):

    mu_gene, gene_idx = _get_mu_gene(X, min_mean)

    tmp = X[:, gene_idx]
    tmp.data **= 2
    var_gene = tmp.mean(axis=0).A.squeeze() - mu_gene ** 2
    del tmp

    filtered_var_gene = var_gene / mu_gene

    return mu_gene, filtered_var_gene, gene_idx


def _calculate_variability_score(
    X, min_mean=0, n_bins=50, fit_percentile=0.1, error_weight=1
):

    """
    Calculate v-score (above-Poisson noise statistic) for genes in the input sparse counts matrix
    Return v-scores and other stats
    """

    mu_gene, filtered_var_gene, gene_idx = _get_variable_gene(X, min_mean)
    data_x = np.log(mu_gene)
    data_y = np.log(filtered_var_gene / mu_gene)

    x, y = _calculate_running_quantile(data_x, data_y, fit_percentile, n_bins)

    g_log = lambda input: np.log(input[1] * np.exp(-input[0]) + input[2])
    hist, bins = np.histogram(np.log(filtered_var_gene[mu_gene > 0]), bins=200)
    bins = bins[:-1] + np.diff(bins) / 2
    max_idx = np.argmax(hist)

    c = np.max((np.exp(bins[max_idx]), 1))

    error_func = lambda b2: np.sum(abs(g_log([x, c, b2]) - y) ** error_weight)
    b = scipy.optimize.fmin(func=error_func, x0=[0.1], disp=False)
    a = c / (1 + b) - 1

    var_scores = filtered_var_gene / ((1 + a) * (1 + b) + b * mu_gene)
    CV_input, CV_eff = np.sqrt((1 + a) * (1 + b) - 1), np.sqrt(b)

    return var_scores, CV_eff, CV_input, gene_idx, mu_gene, filtered_var_gene, a, b


def _make_plot():

    fig = vinplots.Plot()
    fig.construct(nplots=1, ncols=1)
    fig.modify_spines(
        ax="all",
        color="grey",
        spines_to_color=["bottom", "left"],
        spines_positioning_amount=5,
        spines_to_delete=["top", "right"],
        spines_to_move=["bottom", "left"],
    )
    ax = fig.AxesDict[0][0]
    ax.grid(True, zorder=0, alpha=0.25)

    return fig, ax


def _plot_var_score(
    filtered_var_gene,
    mu_gene,
    idx,
    a,
    b,
    sample_name,
    x_label="log10(mean)",
    y_label=("log10(dispersion)"),
):

    fig, ax = _make_plot()

    x_min, x_max = 0.5 * np.min(mu_gene), 2 * np.max(mu_gene)
    x_th = x_min * np.exp(np.log(x_max / x_min) * np.linspace(0, 1, 100))
    y_th = (1 + a) * (1 + b) + b * x_th

    plt.scatter(
        np.log10(mu_gene),
        np.log10(filtered_var_gene),
        c=[[0.8, 0.8, 0.8]],
        alpha=0.3,
        s=3,
        zorder=2,
    )
    plt.scatter(
        np.log10(mu_gene)[idx],
        np.log10(filtered_var_gene)[idx],
        c=[[0, 0, 0]],
        alpha=0.3,
        s=3,
        zorder=3,
    )
    plt.plot(
        np.log10(x_th),
        np.log10(y_th),
        zorder=3,
    )

    plt.title(sample_name, y=1.05)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    plt.show()

def _annotate_adata_highvar_genes(adata, highly_variable_genes_idx):
    
    hv_gene_annotation = np.zeros(adata.X.shape[1])
    hv_gene_annotation[highly_variable_genes_idx] = True
    adata.var["highly_variable"] = hv_gene_annotation.astype(bool)
    
    adata.uns['highly_variable_genes_idx'] = highly_variable_genes_idx        

def _filter_static_genes(
    adata,
    base_idx=[],
    min_var_score_percentile=85,
    min_counts=3,
    min_cells=3,
    plot=True,
    sample_name="Variable genes",
    return_hv_genes=False,
):

    """
    Filter genes by expression level and variability
    Return list of filtered gene indices
    """
    
    n_cells = adata.X.shape[0]
    base_idx = _get_base_idx(base_idx, n_cells)

    (
        var_scores,
        CV_eff,
        CV_input,
        gene_idx,
        mu_gene,
        filtered_var_gene,
        a,
        b,
    ) = _calculate_variability_score(
        adata.X[base_idx, :], min_mean=0, n_bins=50, fit_percentile=0.1, error_weight=1
    )

    gene_idx = gene_idx[var_scores > 0]
    mu_gene = mu_gene[var_scores > 0]
    filtered_var_gene = filtered_var_gene[var_scores > 0]

    min_var_score = np.percentile(var_scores, min_var_score_percentile)
    idx = ((adata.X[:, gene_idx] >= min_counts).sum(0).A.squeeze() >= min_cells) & (
        var_scores >= min_var_score
    )

    if plot:
        _plot_var_score(filtered_var_gene, mu_gene, idx, a, b, sample_name)

    highly_variable_genes_idx = gene_idx[idx]
    n_var_genes = licorice.font_format(str(len(highly_variable_genes_idx)), ["BOLD"])
    print("{} variable genes identified".format(n_var_genes))
    
    _annotate_adata_highvar_genes(adata, highly_variable_genes_idx)
           
    if return_hv_genes:
        return highly_variable_genes_idx



# cell_tools __init__.py

__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# from . import _ATAC as atac
from . import _RNA as rna
# from . import _multiome as multi
# from . import _datasets as data
# from . import _readwrite as io
# from . import _utilities as util
# from . import _general_tools as tl



import pandas as pd
import numpy as np
import statsmodels.sandbox.stats.multicomp
import scipy.stats
    

def _calculate_p_values(X_norm, cluster_cells, compare_cells, gene_idx):

    """"""

    X1 = X_norm[cluster_cells, :][:, gene_idx].toarray()
    X2 = X_norm[compare_cells, :][:, gene_idx].toarray()
    p_values = np.zeros(gene_idx.sum())

    for i, gene in enumerate(np.nonzero(gene_idx)[0]):
        p_values[i] = scipy.stats.ranksums(X1[:, i], X2[:, i])[1]
    p_values = statsmodels.sandbox.stats.multicomp.multipletests(
        p_values,
        alpha=0.05,
        method="fdr_bh",
    )[1]
    del X1, X2

    return p_values


def _calculate_highest_means(
    X_norm, cluster, clusters, cluster_cells, gene_idx, pseudocount
):

    """Calculate highest vs. 2nd-highest means"""

    m1 = X_norm[cluster_cells, :][:, gene_idx].mean(0).A.squeeze()
    m2 = np.zeros(gene_idx.sum())

    for c in np.unique(clusters):
        if c != cluster:
            i_test = clusters == c
            m_test = norm_counts[i_test, :][:, gene_idx].mean(0).A.squeeze()
            gene_idx_change = m_test - m2 > 0
            m2[gene_idx_change] = m_test[gene_idx_change]

    log_fc = np.log2((m1 + pseudocount) / (m2 + pseudocount))

    return log_fc

def _identify_significant_genes(
    results,
    cluster,
    gene_list,
    gene_idx,
    p_values,
    max_pval,
    log_fc,
    min_log_fold_change,
):

    """"""

    qualifying_genes = (p_values < max_pval) & (log_fc > min_log_fold_change)

    if qualifying_genes.sum() > 0:
        results["group"].extend([cluster for i in range(qualifying_genes.sum())])
        results["gene"].extend(list(gene_list[gene_idx][qualifying_genes]))
        results["p-value"].extend(list(p_values[qualifying_genes]))
        results["log2_fold_change"].extend(list(log_fc[qualifying_genes]))

    if verbose:
        print(
            "cluster={}, n_cells={}, n_diff={}".format(
                cluster, cluster_cells.sum(), qualifying_genes.sum()
            )
        )

    results_df = pd.DataFrame(results)[["group", "gene", "p-value", "log2_fold_change"]]

    return results_df

def _find_marker_genes(X_norm, genes, label_column):

    results = {"group": [], "gene": [], "p-value": [], "log2_fold_change": []}
    min_log_fold_change = np.log2(min_fold_change)
    groups = adata.obs[label_column]
    unique_groups = np.unique(adata.obs[label_column])

    for cluster in unique_groups:

        cluster_cells = groups == cluster
        compare_cells = ~cluster_cells

        gene_idx = (X_norm[cluster_cells, :] > 0).sum(
            0
        ).A.squeeze() / cluster_cells.sum() > min_fraction_expressed

        p_values = _calculate_p_values(X_norm, cluster_cells, compare_cells, gene_idx)
        log_fc = _calculate_highest_means(
            X_norm, cluster, clusters, cluster_cells, gene_idx
        )

        results_df = _identify_significant_genes(
            results,
            cluster,
            gene_list,
            gene_idx,
            p_values,
            max_pval,
            log_fc,
            min_log_fold_change,
        )
        return results_df



import scanpy as sc
import numpy as np
import licorice
import matplotlib.pyplot as plt
import vinplots


def _build_plot(geneset_score_key):
    fig = vinplots.Plot()
    fig.construct(
        nplots=5,
        ncols=5,
        width_ratios=[2, 2, 2, 2, 0.05],
        wspace=0.15,
        figsize_width=1,
    )
    fig.modify_spines(
        ax="all",
        spines_to_delete=["top", "right"],
        spines_to_move=["bottom", "left"],
        spines_positioning_amount=15,
    )

    axes = []
    for i in range(len(fig.AxesDict[0])):
        axes.append(fig.AxesDict[0][i])
    titles = ["Unfiltered", "Filtered"]

    axes[1].set_title("Ranked {}".format(geneset_score_key))
    axes[1].set_xlabel("Cells")
    axes[1].set_ylabel("Score")

    for n, ax in enumerate([fig.AxesDict[0][2], fig.AxesDict[0][3]]):
        ax.grid(True, alpha=0.3, zorder=0)
        ax.set_xlabel("UMAP-1")
        ax.set_ylabel("UMAP-2")
        ax.set_title(titles[n], fontsize=16, y=1.03)
        
    fig.AxesDict[0][0].grid(True, alpha=0.3, zorder=0)
    fig.AxesDict[0][1].grid(True, alpha=0.3, zorder=0)
    
    return fig, axes

def _plot_score_distribution(ax, scores, n_bins):

    """"""

    lower = scores.mean() - scores.std() / 2
    upper = scores.mean() + scores.std() / 2
    bh = ax.hist(scores, range=(lower, upper), bins=n_bins, color="darkred")
    
    return lower, upper

def _plot_ranked_scores(adata, score_name, ax):

    """"""

    ranked_scores = adata.obs.sort_values(score_name, ascending=False)[
        score_name
    ].values
    ax.scatter(np.arange(len(ranked_scores)), ranked_scores, s=100, c="black")
    ax.scatter(np.arange(len(ranked_scores)), ranked_scores, s=15, c=ranked_scores)


def _plot_pre_post_filtering(adata, geneset_score_key, score_threshold, n_bins=50):

    """"""
    fig, axes = _build_plot(geneset_score_key)

    vmin, vmax = _plot_score_distribution(axes[0], adata.obs[geneset_score_key], n_bins)
    _plot_ranked_scores(adata, geneset_score_key, ax=axes[1])

    umap = adata.obsm["X_umap"]
    adata_f = adata[adata.obs[geneset_score_key] < score_threshold]
    umap_f = adata_f.obsm["X_umap"]

    im0 = axes[2].scatter(
        umap[:, 0], 
        umap[:, 1], 
        c=adata.obs[geneset_score_key], 
        zorder=10, 
        s=2,
        vmin=vmin,
        vmax=vmax,
    )
    im1 = axes[3].scatter(
        umap_f[:, 0], 
        umap_f[:, 1], 
        c=adata_f.obs[geneset_score_key], 
        zorder=10, 
        s=2,
        vmin=vmin,
        vmax=vmax,
    )
    plt.colorbar(im0, cax=axes[4], orientation="vertical")

def _print_summary(adata, adata_f):

    """"""

    n_cells_init = adata.shape[0]
    n_cells_filt = adata_f.shape[0]
    n_cells_removed = adata.shape[0] - adata_f.shape[0]

    print(
        "Cells before filtering:\t{}".format(
            licorice.font_format(str(n_cells_init), ["BOLD"])
        )
    )
    print(
        "Cells after filtering:\t{} ({} cells removed)".format(
            licorice.font_format(str(n_cells_filt), ["BOLD"]),
            licorice.font_format(str(n_cells_removed), ["BOLD"]),
        )
    )

class _GeneSetFiltering:
    def __init__(self, adata=False, gene_set=False, score_name="gene_set_score"):

        """"""

        if gene_set:
            self._gene_set = gene_set

        if adata:
            self._adata = adata

        self._score_name = score_name

    def plot_gene_set(self, adata=False, gene_set=False):

        if gene_set:
            print("gene set is....")
            self._gene_set = gene_set

        sc.pl.umap(self._adata, color=self._gene_set)

    def score_gene_set(self, adata=False, gene_set=False, score_name=False):

        if score_name:
            self._score_name = score_name

        if gene_set:
            self._gene_set = gene_set
            print("ok gene set is...")

        sc.tl.score_genes(self._adata, self._gene_set, score_name=self._score_name)
        sc.pl.umap(self._adata, color=self._score_name)

    def compare_on_score(self, geneset_score_key=False, score_threshold=0.1):

        if geneset_score_key:
            self._score_name = score_name

        self._score_threshold = score_threshold
        #         self._adata_f = self._adata[self._adata.obs[self._score_name] < self._score_threshold]
        _plot_pre_post_filtering(self._adata, self._score_name, self._score_threshold)

    def filter_on_score(self, score_threshold=0.1):

        if score_threshold != self._score_threshold:
            print(
                "Updating score threshold from {} to {}...".format(
                    self._score_threshold, score_threshold
                )
            )
        self._score_threshold = score_threshold
        adata_f =  self._adata[self._adata.obs[self._score_name] < self._score_threshold]
        _print_summary(self._adata, adata_f)

        return adata_f



# scRNA-seq __init__.py

__module_name__ = "__init__.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])

from ._funcs._calculate_differential_gene_expression import _calculate_differential_gene_expression as DE
from ._funcs._filter_static_genes import _filter_static_genes as filter_static_genes
from ._funcs._find_marker_genes import _find_marker_genes as id_marker_genes

from ._funcs._remove_correlated_genes_adata import _remove_correlated_genes_adata as remove_correlated_genes

from ._funcs._filter_cells_by_gene_set import _GeneSetFiltering as gene_set_filtering


def _filter_cells(dictionary, filter_mask):
    
    """Filter 1- and 2-D entries in a dictionary"""
    
    for key, value in dictionary.items():
        
        if key != "meta":
            if len(value.shape) == 1:
                dictionary[key] = value[filter_mask]
            else:
                dictionary[key] = value[:, filter_mask]
                
    return dictionary


import statsmodels.sandbox.stats.multicomp
import scipy.stats
    
def _calculate_normalized_masked_expression(X, mask):
    return (X[mask, :] > 0).sum(0).A.squeeze() / len(mask)

def _get_gene_mask(X, mask1, mask2, min_fraction_expressed):
    
    """
    
    """
    
    X_1 = _calculate_normalized_masked_expression(X, mask1) > min_fraction_expressed
    X_2 = _calculate_normalized_masked_expression(X, mask2) > min_fraction_expressed
    
    gene_mask = X_1 | X_2
    
    print(X_1.shape, X_2.shape, gene_mask.shape)
    print(X_1.sum(), X_2.sum(), gene_mask.sum())

    return gene_mask

def _calculate_p_values(gene_mask, X1, X2):
    
    """"""
    
    p_values = np.zeros(gene_mask.sum())
    
    for i, gene in enumerate(np.nonzero(gene_mask)[0]):
        p_values[i] = scipy.stats.ranksums(X1[:,i], X2[:,i])[1]
        
    p_values = statsmodels.sandbox.stats.multicomp.multipletests(p_values, alpha=0.05, method='fdr_bh',)[1]
    
    return p_values

def _calculate_differential_gene_expression(adata, mask1, mask2, min_fraction_expressed=0.05, pseudocount=1):
    
    """"""
    
    gene_mask = _get_gene_mask(adata.X, mask1, mask2, min_fraction_expressed)
    
    print(gene_mask.sum())
    X1 = adata.X[mask1,:][:,gene_mask].toarray()
    X2 = adata.X[mask2,:][:,gene_mask].toarray()
    
    m1, m2 = X1.mean(0) + pseudocount, X2.mean(0) + pseudocount
    r = np.log2(m1 / m2)
    
    p_values = _calculate_p_values(gene_mask, X1, X2)

    df = pd.DataFrame({
            'gene': adata.var_names.values.astype(str)[gene_mask],
            'pv': p_values,
            'm1': m1 - pseudocount,
            'm2': m2 - pseudocount,
            'ratio': r
        })

    return df



# setup.py

__module_name__ = "setup.py"
__author__ = ", ".join(["Michael E. Vinyard"])
__email__ = ", ".join(["vinyard@g.harvard.edu",])


# package imports #
# --------------- #
import os
import re
from setuptools import setup
import sys


setup(
    name="cell_tools",
    version="0.0.21",
    python_requires=">3.7.0",
    author="Michael E. Vinyard - Harvard University - Massachussetts General Hospital - Broad Institute of MIT and Harvard",
    author_email="mvinyard@broadinstitute.org",
    url="https://github.com/mvinyard/cell-tools",
    long_description=open("README.md", encoding="utf-8").read(),
    long_description_content_type="text/markdown",
    description="cell_tools - Analysis Tools for Single-Cell Data",
    packages=[
        "cell_tools",
        "cell_tools._RNA",
        "cell_tools._RNA._funcs",
    ],
    install_requires=[
        "anndata>=0.7.8",
        "cython>=0.29.24",
	"episcanpy>=0.3.2",
	"nb-black>=1.0.7",
        "pyranges>=0.0.113",
	"licorice>=0.0.2",
	"pydk>=0.0.4",
	"vinplots>=0.0.43",
	"tqdm>=4.62.3",
    ],
    classifiers=[
        "Development Status :: 2 - Pre-Alpha",
        "Programming Language :: Python :: 3.7",
        "Intended Audience :: Science/Research",
        "Topic :: Scientific/Engineering :: Bio-Informatics",
    ],
    license="MIT",
)
